# The RMVPE model is from https://github.com/Dream-High/RMVPE
# Apache License 2.0: https://github.com/Dream-High/RMVPE/blob/main/LICENSE
# With modifications from https://github.com/RVC-Project/Retrieval-based-Voice-Conversion-WebUI/blob/main/infer/lib/rmvpe.py
# MIT License: https://github.com/RVC-Project/Retrieval-based-Voice-Conversion-WebUI/blob/main/LICENSE

from .model import RMVPE
